import logging

from mongoengine import signals
from nefertari.utils import to_dicts


log = logging.getLogger(__name__)


def on_post_save(sender, document, **kw):
    """ Add new document to index or update existing. """
    from nefertari.elasticsearch import ES
    common_kw = {'request': getattr(document, '_request', None)}
    created = kw.get('created', False)
    if created:
        es = ES(document.__class__.__name__)
        es.index(document.to_dict(), **common_kw)
    elif not created and document._get_changed_fields():
        es = ES(document.__class__.__name__)
        es.index(document.to_dict(), **common_kw)
        es.index_relations(document, nested_only=True, **common_kw)


def on_post_delete(sender, document, **kw):
    from nefertari.elasticsearch import ES
    request = getattr(document, '_request', None)
    ES(document.__class__.__name__).delete(
        document.id, request=request)


def on_bulk_update(model_cls, objects, request):
    if not getattr(model_cls, '_index_enabled', False):
        return

    if not objects:
        return

    from nefertari.elasticsearch import ES
    es = ES(source=model_cls.__name__)
    documents = to_dicts(objects)
    es.index(documents, request=request)

    # Reindex relationships
    es.bulk_index_relations(objects, request=request, nested_only=True)


def setup_es_signals_for(source_cls):
    signals.post_save.connect(on_post_save, sender=source_cls)
    signals.post_delete.connect(on_post_delete, sender=source_cls)
    log.info('setup_es_signals_for: %r' % source_cls)



import logging

import mongoengine

from .documents import (
    BaseDocument, ESBaseDocument, BaseMixin,
    get_document_cls, get_document_classes)
from .serializers import JSONEncoder, ESJSONSerializer
from .metaclasses import ESMetaclass
from .utils import (
    relationship_fields, is_relationship_field,
    get_relationship_cls)
from .fields import (
    BigIntegerField,
    BooleanField,
    DateField,
    DateTimeField,
    ChoiceField,
    FloatField,
    IntegerField,
    IntervalField,
    BinaryField,
    DecimalField,
    PickleField,
    SmallIntegerField,
    StringField,
    TextField,
    TimeField,
    UnicodeField,
    UnicodeTextField,
    Relationship,
    IdField,
    ForeignKeyField,
    ListField,
    DictField,
)

__all__ = [
    'BigIntegerField',
    'BooleanField',
    'DateField',
    'DateTimeField',
    'ChoiceField',
    'FloatField',
    'IntegerField',
    'IntervalField',
    'BinaryField',
    'DecimalField',
    'PickleField',
    'SmallIntegerField',
    'StringField',
    'TextField',
    'TimeField',
    'UnicodeField',
    'UnicodeTextField',
    'Relationship',
    'IdField',
    'ForeignKeyField',
    'ListField',
    'DictField',
    'BaseDocument',
    'ESBaseDocument',
    'BaseMixin',
    'get_document_cls',
    'get_document_classes',
    'relationship_fields',
    'is_relationship_field',
    'get_relationship_cls',
    'JSONEncoder',
    'ESJSONSerializer',
    'ESMetaclass',
    'setup_database',
    ]

log = logging.getLogger(__name__)


def includeme(config):
    """ Include required packages. """
    pass


def setup_database(config):
    """ Setup db engine and db itself. Create db if it doesn't exist. """
    settings = config.registry.settings
    mongoengine.connect(settings['mongodb.db'],
                        host=settings['mongodb.host'],
                        port=int(settings['mongodb.port']))



import datetime
import pickle
from functools import partial

import six
import dateutil.parser
import mongoengine as mongo
from mongoengine import fields
from mongoengine.queryset import DO_NOTHING, NULLIFY, CASCADE, DENY, PULL


class BaseFieldMixin(object):
    """ Base mixin to implement a common interface for all mongo fields.

    It is responsible for dropping invalid field kwargs.
    Subclasses must define `translate_kwargs` if some fields from
    input kwargs have different names but the same meaning.

    Attributes:
        _valid_kwargs: sequence of strings that represent valid kwargs
            names that a current field may accept. Kwargs not present in
            this sequence are dropped by `self.drop_invalid_kwargs`.
        _common_valid_kwargs: sequence of strings that represent valid kwargs
            names that a `mongoengine.base.fields.BaseField` accepts. These
            kwargs are common to all the fields.
    """
    _valid_kwargs = ()
    _common_valid_kwargs = (
        'db_field', 'name', 'required', 'default', 'unique',
        'unique_with', 'primary_key', 'validation', 'choices',
        'verbose_name', 'help_text', 'sparse', 'onupdate',
        # 'null'  # In development branch of mongoengine
    )

    def __init__(self, *args, **kwargs):
        """ Translate kwargs and drop invalid kwargs. """
        self._init_kwargs = kwargs.copy()
        kwargs = self.translate_kwargs(kwargs)
        kwargs = self.drop_invalid_kwargs(kwargs)
        self.onupdate = kwargs.pop('onupdate', None)
        super(BaseFieldMixin, self).__init__(*args, **kwargs)

    def clean(self, instance):
        """ Perform field clean.

        In particular:
            - On update set field value to `self.onupdate`

        Arguments:
            :instance: BaseDocument instance to which field is attached
        """
        if self.onupdate is not None and not instance._created:
            value = self.onupdate
            if six.callable(value):
                value = value()
            setattr(instance, self.name, value)

    def translate_kwargs(self, kwargs):
        """ Translate kwargs from one key to another.

        E.g. your field is an instance of a StringField and accepts kwarg
        'maximum_length', but the StringField expects the same kwargs with
        a key `max_length`. Extend this method to perform such translations.

        This particular implementation translates:
            name -> db_field
        """
        kwargs = kwargs.copy()
        kwargs['db_field'] = kwargs.pop('name', None)
        # In development branch of mongoengine
        # if kwargs.get('null') is None:
        #     kwargs['null'] = not kwargs.get('required', False)
        return kwargs

    def drop_invalid_kwargs(self, kwargs):
        """ Drop kwargs not present in `self._common_valid_kwargs` and
        `self._valid_kwargs` (thus invalid).
        """
        valid_kwargs = self._common_valid_kwargs + self._valid_kwargs
        return {k: v for k, v in kwargs.items() if k in valid_kwargs}


class IntegerField(BaseFieldMixin, fields.IntField):
    _valid_kwargs = ('min_value', 'max_value')


class BigIntegerField(BaseFieldMixin, fields.LongField):
    _valid_kwargs = ('min_value', 'max_value')


class SmallIntegerField(BaseFieldMixin, fields.IntField):
    """ As mongoengine does not provide SmallInt, this is just a
    copy of `IntegerField`.
    """
    _valid_kwargs = ('min_value', 'max_value')


class BooleanField(BaseFieldMixin, fields.BooleanField):
    _valid_kwargs = ()


class DateField(BaseFieldMixin, fields.DateTimeField):
    """ Custom field that stores `datetime.date` instances.

    This is basically mongoengine's `DateTimeField` which gets
    datetime's `.date()` and formats it into a string before storing
    to mongo.
    """
    _valid_kwargs = ()

    def validate(self, value):
        new_value = self.to_python(value)
        if not isinstance(new_value, datetime.date):
            self.error("Can't parse date `%s`" % value)

    def to_python(self, *args, **kwargs):
        value = super(DateField, self).to_python(*args, **kwargs)
        if isinstance(value, six.string_types):
            value = dateutil.parser.parse(value, dayfirst=False).date()
        return value

    def to_mongo(self, *args, **kwargs):
        """ Override mongo's DateTimeField value conversion to get
        date instead of datetime.
        """
        value = super(DateField, self).to_mongo(*args, **kwargs)
        if isinstance(value, datetime.datetime):
            value = value.date()
        return value.strftime('%Y-%m-%d')


class DateTimeField(BaseFieldMixin, fields.DateTimeField):
    _valid_kwargs = ()


class FloatField(BaseFieldMixin, fields.FloatField):
    _valid_kwargs = ('min_value', 'max_value')

    def validate(self, value):
        """ Override to try to convert string values to floats. """
        if isinstance(value, six.string_types):
            try:
                value = float(value)
            except ValueError:
                pass
        return super(FloatField, self).validate(value)


class StringField(BaseFieldMixin, fields.StringField):
    _valid_kwargs = ('regex', 'min_length', 'max_length')


class TextField(StringField):
    pass


class UnicodeField(BaseFieldMixin, fields.StringField):
    _valid_kwargs = ('regex', 'min_length', 'max_length')


class UnicodeTextField(UnicodeField):
    pass


class ChoiceField(fields.BaseField):
    """
    As mongoengine does not have an explicit ChoiceField, but all mongoengine
    fields accept `choices` kwarg, we need to define a proxy here.
    It uses a naive aproach: check the type of first choice and instantiate
    a field of an appropriate type under the hood. Then translate all the
    attribute access to the underlying field.
    """
    def __init__(self, *args, **kwargs):
        self._init_kwargs = kwargs.copy()
        first_choice = kwargs['choices'][0]
        if isinstance(first_choice, int):
            self._real_field = IntegerField(*args, **kwargs)
        elif isinstance(first_choice, six.text_type):
            self._real_field = UnicodeField(*args, **kwargs)
        elif isinstance(first_choice, six.string_types):
            self._real_field = StringField(*args, **kwargs)
        elif isinstance(first_choice, float):
            self._real_field = FloatField(*args, **kwargs)
        else:
            raise ValueError(
                'Choices must be one of the following types: int, str, float, '
                'unicode. All choices must be of one type.')
        self.__dict__.update(self._real_field.__dict__)

    def __getattribute__(self, attr):
        methods = {'to_python', 'to_mongo', 'prepare_query_value',
                   'validate', 'lookup_member'}
        if attr in methods:
            return self._real_field.__getattribute__(attr)
        return super(ChoiceField, self).__getattribute__(attr)


class BinaryField(BaseFieldMixin, fields.BinaryField):
    _valid_kwargs = ('max_bytes',)

    def translate_kwargs(self, kwargs):
        """ Translate kwargs from one key to another.

        Translates:
            length -> max_bytes
        """
        kwargs = super(BinaryField, self).translate_kwargs(kwargs)
        kwargs['max_bytes'] = kwargs.pop('length', None)
        return kwargs


class DecimalField(BaseFieldMixin, fields.DecimalField):
    """ This is basically a DecimalField with a fixed name of
    `precision` kwarg.
    """
    _valid_kwargs = ('min_value', 'max_value', 'force_string',
                     'precision', 'rounding')

    def translate_kwargs(self, kwargs):
        """ Translate kwargs from one key to another.

        Translates:
            scale -> precision
        """
        kwargs = super(DecimalField, self).translate_kwargs(kwargs)
        kwargs['precision'] = kwargs.pop('scale')
        return kwargs


class TimeField(BaseFieldMixin, fields.BaseField):
    """ Custom field that stores `datetime.date` instances. """
    _valid_kwargs = ()

    def validate(self, value):
        new_value = self.to_python(value)
        if not isinstance(new_value, datetime.time):
            self.error("Can't parse time `%s`" % value)

    def to_mongo(self, value):
        value = super(TimeField, self).to_mongo(value)
        if not isinstance(value, six.string_types):
            value = value.strftime('%H:%M:%S')
        return value

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.time):
            return value
        if isinstance(value, datetime.datetime):
            return value.time()
        if six.callable(value):
            return value()

        if not isinstance(value, six.string_types):
            return None

        try:
            return dateutil.parser.parse(value).time()
        except (TypeError, ValueError):
            return None

    def prepare_query_value(self, op, value):
        return self.to_mongo(value)


class PickleField(BinaryField):
    """ Custom field that stores pickled data as a BinaryField.

    Data is pickled when saving to mongo and unpickled when retrieving
    from mongo.
    The `pickler` kwarg may be provided that may reference any object with
    pickle-compatible `dumps` and `loads` methods. Defaults to python's
    built-in `pickle`.
    """
    pickler = None

    def translate_kwargs(self, kwargs):
        self.pickler = kwargs.pop('pickler', pickle)
        return super(PickleField, self).translate_kwargs(kwargs)

    def validate(self, value):
        value = self.pickler.dumps(value)
        return super(PickleField, self).validate(value)

    def to_mongo(self, value):
        value = self.pickler.dumps(value)
        return super(PickleField, self).to_mongo(value)

    def to_python(self, value):
        value = super(PickleField, self).to_python(value)
        return self.pickler.loads(value)


class IntervalField(IntegerField):
    """ Custom field that stores `datetime.timedelta` instances.

    Values are stored as seconds in mongo and loaded by
    `datetime.timedelta(seconds=<value>) when restoring from mongo.
    """
    _valid_kwargs = ()

    def validate(self, value):
        value = self.to_mongo(value)
        return super(IntervalField, self).validate(value)

    def to_mongo(self, value):
        if isinstance(value, datetime.timedelta):
            value = int(value.total_seconds())
        return value

    def to_python(self, value):
        return datetime.timedelta(seconds=value)

    def prepare_query_value(self, op, value):
        return self.to_mongo(value)


class ListField(BaseFieldMixin, fields.ListField):
    """ Custom ListField.

    The custom part is the validation. Choices are stored in a separate
    attribute :self.list_choices: and validation checks if the value (which is
    a sequence) contains anything other than the choices specified in
    :self.list_choices:.

    The original mongoengine ListField validation requires the value to be a
    sequence but checks for value(sequence) inclusion in choices.
    """
    _valid_kwargs = ('field',)

    def __init__(self, *args, **kwargs):
        self.list_choices = kwargs.pop('choices', None)
        self.item_type = kwargs.pop('item_type')
        super(ListField, self).__init__(*args, **kwargs)
        self._init_kwargs.update(
            list_choices=self.list_choices,
            item_type=self.item_type)

    def validate(self, value, **kwargs):
        super(ListField, self).validate(value, **kwargs)
        if self.list_choices and value is not None:
            choice_list = self.list_choices
            if isinstance(self.list_choices[0], (list, tuple)):
                choice_list = [k for k, v in self.list_choices]
            if set(value) - set(choice_list):
                self.error('Value must be one of {}. Got: {}'.format(
                    str(choice_list), str(value)))
        return value


class DictField(BaseFieldMixin, fields.DictField):
    _valid_kwargs = ('basecls', 'field')


class IdField(BaseFieldMixin, fields.ObjectIdField):
    """ Just a subclass of ObjectIdField that must be used for fields
    that represent database-specific 'id' field.
    """
    _valid_kwargs = ('primary_key',)

    def __init__(self, *args, **kwargs):
        kwargs.pop('primary_key', None)
        super(IdField, self).__init__(*args, **kwargs)


class ForeignKeyField(BaseFieldMixin, fields.StringField):
    """ Field that references another document.

    It is not meant to be used inside the mongodb engine. It is added for
    compatibility with sqla and is not displayed in JSON output.
    """
    _valid_kwargs = ()


class ReferenceField(BaseFieldMixin, fields.ReferenceField):
    """ Field that references another document.

    **It ISN'T MEANT to be used explicitly by the user. To create a
    relationship field, use the `Relationship` constructor function.**

    When this field is not added to a model's `_nested_relationships`, this
    field returns an ID of the document that is being referenced. Otherwise
    the full document is included in JSON response.

    This class is used in a `Relationship` function to generate a kind of
    one-to-one relationship. It is also used to create backreferences.

    `reverse_rel_field`: string name of a field on the related document.
        Used when generating backreferences so that fields on each side
        know the name of the field on the other side.
    """
    _valid_kwargs = (
        'document_type', 'dbref', 'reverse_delete_rule',
    )
    _kwargs_prefix = 'ref_'
    _backref_prefix = 'backref_'
    reverse_rel_field = None

    def __init__(self, *args, **kwargs):
        """ Init the field.

        Also saves backref kwargs for future creation of the backref.

        Expects:
            `document` or `<_kwargs_prefix>document`: mongoengine model name.
        """
        _init_kwargs = kwargs.copy()
        backref_prefix_len = len(self._backref_prefix)
        key = 'document'
        pref_key = self._kwargs_prefix + key
        document_type = kwargs.pop(pref_key, None) or kwargs.pop(key, None)
        args = (document_type,) + args

        # Filter out backreference kwargs
        self.backref_kwargs = {
            k[backref_prefix_len:]: v for k, v in kwargs.items()
            if k.startswith(self._backref_prefix)}

        super(ReferenceField, self).__init__(*args, **kwargs)
        self._init_kwargs = _init_kwargs

    def _register_deletion_hook(self, old_object, instance):
        """ Register a backref hook to delete the `instance` from the
        `old_object`'s field to which the `instance` was related before
        by the backref.

        `instance` is either deleted from the `old_object` field's collection
        or `old_object`'s field, responsible for relationship is set to None.
        This depends on type of the field at `old_object`.

        `instance` is not actually used in hook - up-to-date value of
        `instance` is passed to hook when it is run.
        """
        def _delete_from_old(old_obj, document, field_name):
            from mongoengine.fields import ListField
            field_value = getattr(old_obj, field_name, None)
            if field_value:
                field_object = old_obj._fields[field_name]
                if isinstance(field_object, ListField):
                    new_value = list(field_value or [])
                    if document in new_value:
                        new_value.remove(document)
                else:
                    new_value = (None if field_value == document
                                 else field_value)

                if new_value != field_value:
                    old_obj.update({field_name: new_value})

        old_object_hook = partial(
            _delete_from_old,
            old_obj=old_object,
            field_name=self.reverse_rel_field)
        instance._backref_hooks += (old_object_hook,)

    def _register_addition_hook(self, new_object, instance):
        """ Register backref hook to add `instance` to the `new_object`s
        field to which `instance` became related by the backref.

        `instance` is either added to `new_object` field's collection or
        `new_object`s field, responsible for relationship is set to `instance`.
        This depends on type of the field at `new_object`.

        `instance` is not actually used in hook - up-to-date value of
        `instance` is passed to hook when it is run.
        """
        def _add_to_new(new_obj, document, field_name):
            from mongoengine.fields import ListField
            field_value = getattr(new_obj, field_name, None)
            field_object = new_obj._fields[field_name]
            if isinstance(field_object, ListField):
                new_value = list(field_value or [])
                if document not in new_value:
                    new_value.append(document)
            else:
                new_value = (document if field_value != document
                             else field_value)

            if new_value != field_value:
                new_obj.update({field_name: new_value})

        new_object_hook = partial(
            _add_to_new,
            new_obj=new_object,
            field_name=self.reverse_rel_field)
        instance._backref_hooks += (new_object_hook,)

    def __set__(self, instance, value):
        """ Custom __set__ method that updates linked relationships.

        Updates linked relationship fields if the current field has a
        `reverse_rel_field` property set. By default this property is
        only set when the backreference is created.

        If value is changed, `instance` is deleted from object it was related
        to before and is added to object it will be related to now - `value`.
        """
        super_set = super(ReferenceField, self).__set__

        # Object has no backref / is not backref
        if not self.reverse_rel_field:
            return super_set(instance, value)

        old_object = getattr(instance, self.name)
        new_object = value
        super_set(instance, value)
        # The same object is being set - no changes needed
        if new_object == old_object:
            return

        # Old object is being changed to new one, thus
        # old should forget about the `instance`
        if old_object is not None and isinstance(old_object, mongo.Document):
            self._register_deletion_hook(old_object, instance)

        # New object is being set, thus se need to set `instance` to it
        if new_object is not None and isinstance(new_object, mongo.Document):
            self._register_addition_hook(new_object, instance)

    def _get_referential_action(self, kwargs):
        """ Determine/translate generic rule name to mongoengine-specific rule.

        Custom rule names are used here to make them look SQL-ish and
        pretty at the same time.

        Mongoengine rules are:
            DO_NOTHING   Don't do anything (default)
            NULLIFY      Updates the reference to null
            CASCADE      Deletes the documents associated with the reference
            DENY         Prevent the deletion of the reference object
            PULL         Pull the reference from a ListField of references
        """
        key = 'ondelete'
        ondelete = kwargs.pop(key, 'DO_NOTHING')
        rules = {
            'DO_NOTHING': DO_NOTHING,
            'NULLIFY': NULLIFY,
            'CASCADE': CASCADE,
            'RESTRICT': DENY,
            'PULL': PULL,
        }
        ondelete = ondelete.upper()
        if ondelete not in rules:
            raise KeyError('Invalid `{}` argument value. Must be '
                           'one of: {}'.format(key, ', '.join(rules.keys())))
        return rules[ondelete]

    def translate_kwargs(self, kwargs):
        kwargs = super(ReferenceField, self).translate_kwargs(kwargs)
        # Remove prefixes
        for key in self._valid_kwargs:
            pre_key = self._kwargs_prefix + key
            if pre_key in kwargs:
                kwargs[key] = kwargs.pop(pre_key)
        kwargs['reverse_delete_rule'] = self._get_referential_action(kwargs)
        return kwargs


class RelationshipField(BaseFieldMixin, fields.ListField):
    """ Relationship field meant to be used to create one-to-many relationships.

    **It ISN'T MEANT to be used by users explicitly. To create a
    relationship field, use the `Relationship` constructor function.**

    It is used in the `Relationship` function to generate one-to-many
    relationships. Under the hood it is just a ListField containing
    ReferenceFields.

    `reverse_rel_field`: string name of a field on the related document.
        Used when generating backreferences so that fields on each side
        know the name of the field on the other side.
    """
    _valid_kwargs = ('field',)
    _common_valid_kwargs = (
        'db_field', 'required', 'default', 'unique',
        'unique_with', 'primary_key', 'validation', 'choices',
        'verbose_name', 'help_text', 'sparse',
    )

    _backref_prefix = 'backref_'
    reverse_rel_field = None

    def __init__(self, *args, **kwargs):
        """ Save backref kwargs for future creation of backref. """
        # Filter out backreference kwargs
        self.backref_kwargs = {
            k[len(self._backref_prefix):]: v for k, v in kwargs.items()
            if k.startswith(self._backref_prefix)}
        super(RelationshipField, self).__init__(*args, **kwargs)

    def _register_addition_hook(self, new_object, instance):
        """ Define and register addition hook.

        Hook sets `instance` at `new_object` using `self.reverse_rel_field` as
        a field which should be set. `instance` is not actually used in hook -
        up-to-date value of `instance` is passed to hook when it is run.
        """
        def _add_to_new(new_obj, document, field_name):
            field_value = getattr(new_obj, field_name, None)
            if field_value != document:
                new_obj.update({field_name: document})

        new_object_hook = partial(
            _add_to_new,
            new_obj=new_object,
            field_name=self.reverse_rel_field)
        instance._backref_hooks += (new_object_hook,)

    def _register_deletion_hook(self, old_object, instance):
        """ Define and register deletion hook.

        Hook removes `instance` from `new_object` using
        `self.reverse_rel_field` as a field from which value should be removed.
        `instance` is not actually used in hook - up-to-date value of
        `instance` is passed to hook when it is run.
        """
        def _delete_from_old(old_obj, document, field_name):
            field_value = getattr(old_obj, field_name, None)
            if field_value == document:
                old_obj.update({field_name: None})

        old_object_hook = partial(
            _delete_from_old,
            old_obj=old_object,
            field_name=self.reverse_rel_field)
        instance._backref_hooks += (old_object_hook,)

    def __set__(self, instance, value):
        """ Custom __set__ method that updates linked relationships.

        Updates linked relationship fields if current field has
        `reverse_rel_field` property set. By default this property is
        only set when a backreference is created.

        If the value is changed, `instance` is deleted from the object it was
        related to before and is added to the object it will be related to now
        - `value`.
        """
        super_set = super(RelationshipField, self).__set__

        if not self.reverse_rel_field:
            return super_set(instance, value)

        current_value = getattr(instance, self.name, []) or []
        value = value or []
        super_set(instance, value)

        if value == current_value:
            return

        is_doc = lambda x: isinstance(x, mongo.Document)

        added_values = set(value) - set(current_value)
        if added_values and is_doc(list(added_values)[0]):
            for val in added_values:
                self._register_addition_hook(val, instance)

        deleted_values = set(current_value) - set(value)
        if deleted_values and is_doc(list(deleted_values)[0]):
            for val in deleted_values:
                self._register_deletion_hook(val, instance)

    def translate_kwargs(self, kwargs):
        """ For RelationshipField use ReferenceField keys without prefix. """
        kwargs = super(RelationshipField, self).translate_kwargs(kwargs)
        reference_kwargs = {}
        # Extract reference kwargs withour prefix
        fields = ReferenceField._valid_kwargs + ('document', 'ondelete')
        for key in fields:
            if key in kwargs:
                reference_kwargs[key] = kwargs.pop(key)
        reference_field = ReferenceField(**reference_kwargs)
        kwargs['field'] = reference_field
        return kwargs


def Relationship(**kwargs):
    """ Relationship field generator.

    Should be used to generate one-to-many and one-to-one relationships.
    Provide `uselist` to indicate which kind of relation you expect to get.
    If `uselist` is True, then RelationshipField is used and a one-to-many
    is created. Otherwise ReferenceField is used and a one-to-one is created.

    This is the place where `ondelete` rules kwargs should be passed.
    If you switched from the SQLA engine, copy here the same `ondelete` rules
    you passed to SQLA's `ForeignKeyField`.
    `ondelete` kwargs may be kept in both fields with no side-effects when
    switching between the sqla and mongo engines.
    """
    uselist = kwargs.pop('uselist', True)
    field_cls = RelationshipField if uselist else ReferenceField
    return field_cls(**kwargs)



import logging
import datetime
import decimal

import elasticsearch
from bson import ObjectId, DBRef

from nefertari.renderers import _JSONEncoder


log = logging.getLogger(__name__)


class JSONEncoderMixin(object):
    def default(self, obj):
        if isinstance(obj, (ObjectId, DBRef)):
            return str(obj)
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.strftime("%Y-%m-%dT%H:%M:%SZ")  # iso
        if isinstance(obj, datetime.time):
            return obj.strftime('%H:%M:%S')
        if isinstance(obj, datetime.timedelta):
            return obj.seconds
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return super(JSONEncoderMixin, self).default(obj)


class JSONEncoder(JSONEncoderMixin, _JSONEncoder):
    """ JSON encoder class to be used in views to encode response. """
    def default(self, obj):
        if hasattr(obj, 'to_dict'):
            # If it got to this point, it means its a nested object.
            # outter objects would have been handled with DataProxy.
            return obj.to_dict()
        return super(JSONEncoder, self).default(obj)


class ESJSONSerializer(JSONEncoderMixin,
                       elasticsearch.serializer.JSONSerializer):
    """ JSON encoder class used to serialize data before indexing
    to ES. """
    def default(self, obj):
        try:
            return super(ESJSONSerializer, self).default(obj)
        except:
            import traceback
            log.error(traceback.format_exc())



from mongoengine import Document
from mongoengine.queryset import DO_NOTHING

from .signals import setup_es_signals_for
from .fields import ReferenceField, RelationshipField


class DocumentMetaclass(Document.my_metaclass):
    """ Custom metaclass that supports backreferences generation.

    The feature of this metaclass is that it creates a backreference
    for `ReferenceField` and `RelationshipField`(which is a ListField of
    ReferenceFields internally). Backrefs are created for the 'other'(backref)
    side the of relationship when the document class at the 'origin side'
    (the side which defines the original relationship) is being created.

    A backreference is created if the field defining a relationship provides
    an additional set of arguments for backreference that are prefixed with
    `ReferenceField._backref_prefix`. Arguments are the same as for
    `Relationship`, except:
        1. `document` is not a valid argument for backref;
        2. The `name` backref argument must be provided and should be the name
            of the backreference field that will be created on the other side
            of relationship.

    Follow inline comments in the code to understand how the process of backref
    creation works. Check `mongoengine/base/metaclasses.py` for the original
    code of this metaclass.
    """

    def __init__(self, name, bases, attrs):
        """ Override new class initialization to create backreferences.

        """
        super(DocumentMetaclass, self).__init__(name, bases, attrs)
        for field_name, field in self._fields.items():

            # Field is not a relationship field
            if not isinstance(field, (ReferenceField, RelationshipField)):
                continue

            # Field has no backreference kwargs, thus does not use
            # backreference
            if not (hasattr(field, 'backref_kwargs') and field.backref_kwargs):
                continue

            # Prepare kwargs for new ReferenceField
            backref_kw = field.backref_kwargs.copy()
            backref_name = backref_kw.pop('name')
            backref_kw['document'] = self.__name__

            # Create backref ReferenceField. Set its name and `db_field` prop
            backref_field = ReferenceField(**backref_kw)
            backref_field.name = backref_name
            if not backref_field.db_field:
                backref_field.db_field = backref_name

            # Set field's `owner_document`
            if isinstance(field, RelationshipField):
                target_cls = field.field.document_type
            else:
                target_cls = field.document_type
            backref_field.owner_document = target_cls

            # Set `reverse_rel_field` for both fields, so they know name of the
            # field that is used on the other side of relationship
            backref_field.reverse_rel_field = field_name
            field.reverse_rel_field = backref_name

            # Add field to target class
            target_cls._fields[backref_name] = backref_field

            # Add new field to `_fields_ordered`
            if (backref_name in target_cls._fields and
                    backref_name not in target_cls._fields_ordered):
                fields = list(target_cls._fields_ordered) + [backref_name]
                target_cls._fields_ordered = sorted(fields)

            # Set new field as an attribute of target class
            setattr(target_cls, backref_name, backref_field)

            # Register reverse deletion rules
            delete_rule = getattr(backref_field, 'reverse_delete_rule',
                                  DO_NOTHING)
            if delete_rule != DO_NOTHING:
                self.register_delete_rule(
                    target_cls,
                    backref_name,
                    delete_rule)


class ESMetaclass(DocumentMetaclass):
    def __init__(self, name, bases, attrs):
        self._index_enabled = True
        setup_es_signals_for(self)
        return super(ESMetaclass, self).__init__(name, bases, attrs)



import pytest
from mock import patch, Mock

import mongoengine as mongo
from mongoengine.errors import FieldDoesNotExist
from nefertari.utils.dictset import dictset
from nefertari.json_httpexceptions import JHTTPBadRequest

from .. import documents as docs
from .. import fields


class TestDocumentHelpers(object):

    @patch.object(docs.mongo.document, 'get_document')
    def test_get_document_cls(self, mock_get):
        mock_get.return_value = 'foo'
        assert docs.get_document_cls('MyModel') == 'foo'
        mock_get.assert_called_once_with('MyModel')

    @patch.object(docs.mongo.base, 'common')
    def test_get_document_classes(self, mock_common):
        bar_mock = Mock(_meta={'abstract': False})
        mock_common._document_registry = {
            'Foo': Mock(_meta={'abstract': True}),
            'Bar': bar_mock,
            'Zoo': Mock(_meta={}),
        }
        document_classes = docs.get_document_classes()
        assert document_classes == {'Bar': bar_mock}

    @patch.object(docs.mongo.document, 'get_document')
    def test_get_document_cls_error(self, mock_get):
        mock_get.side_effect = Exception()
        with pytest.raises(ValueError) as ex:
            docs.get_document_cls('MyModel')
        mock_get.assert_called_once_with('MyModel')
        assert str(ex.value) == '`MyModel` does not exist in mongo db'

    def test_process_lists(self):
        test_dict = dictset(
            id__in='1,   2, 3',
            name__all='foo',
            other__arg='4',
            yet_other_arg=5,
        )
        result_dict = docs.process_lists(test_dict)
        expected = dictset(
            id__in=['1', '2', '3'],
            name__all=['foo'],
            other__arg='4',
            yet_other_arg=5,
        )
        assert result_dict == expected

    def test_process_bools(self):
        test_dict = dictset(
            complete__bool='false',
            other_arg=5,
        )
        result_dict = docs.process_bools(test_dict)
        assert result_dict == dictset(complete=False, other_arg=5)


class TestBaseMixin(object):

    @patch('nefertari.elasticsearch.engine')
    def test_get_es_mapping(self, mock_conv):
        class MyModel(docs.BaseDocument):
            _nested_relationships = ['parent']
            _nesting_depth = 0
            my_id = fields.IdField()
            name = fields.StringField(primary_key=True)
            status = fields.ChoiceField(choices=['active'])
            groups = fields.ListField(item_type=fields.IntegerField)

        class MyModel2(docs.BaseDocument):
            _nested_relationships = ['child']
            _nesting_depth = 1
            name = fields.StringField(primary_key=True)
            child = fields.Relationship(
                document='MyModel', backref_name='parent',
                uselist=False, backref_uselist=False)

        mymodel_mapping = MyModel.get_es_mapping()
        assert mymodel_mapping == {
            'MyModel': {
                'properties': {
                    '_pk': {'type': 'string'},
                    '_version': {'type': 'long'},
                    'groups': {'type': 'long'},
                    'my_id': {'type': 'string'},
                    'name': {'type': 'string'},
                    'parent': {'type': 'string'},
                    'status': {'type': 'string'},
                }
            }
        }

        mymodel2_mapping = MyModel2.get_es_mapping()
        child_props = mymodel_mapping['MyModel']['properties']
        assert mymodel2_mapping == {
            'MyModel2': {
                'properties': {
                    '_pk': {'type': 'string'},
                    '_version': {'type': 'long'},
                    'name': {'type': 'string'},
                    'child': {
                        'type': 'nested',
                        'properties': child_props
                    },
                }
            }
        }

    def test_pk_field(self):
        class MyModel(docs.BaseDocument):
            my_id = fields.IdField()
            name = fields.StringField(primary_key=True)

        assert MyModel.pk_field() == 'name'

    def test_check_fields_allowed_not_existing_field(self):
        class MyModel(docs.BaseDocument):
            name = fields.StringField()

        with pytest.raises(JHTTPBadRequest) as ex:
            MyModel.check_fields_allowed(('id__in', 'name', 'description'))
        assert "'MyModel' object does not have fields" in str(ex.value)
        assert 'description' in str(ex.value)
        assert 'name' not in str(ex.value)

    def test_check_fields_allowed(self):
        class MyModel(docs.BaseDocument):
            name = fields.StringField()
        try:
            MyModel.check_fields_allowed(('id__in', 'name'))
        except JHTTPBadRequest:
            raise Exception('Unexpected JHTTPBadRequest exception raised')

    def test_check_fields_allowed_dymanic_doc(self):
        class MyModel(docs.BaseMixin, mongo.DynamicDocument):
            name = fields.StringField()
        try:
            MyModel.check_fields_allowed(('id__in', 'name', 'description'))
        except JHTTPBadRequest:
            raise Exception('Unexpected JHTTPBadRequest exception raised')

    def test_filter_fields(self):
        class MyModel(docs.BaseDocument):
            name = fields.StringField()
        params = MyModel.filter_fields(dictset(
            description='nice',
            name='regular name',
            id__in__here=[1, 2, 3],
        ))
        assert params == dictset(
            name='regular name',
            id__in__here=[1, 2, 3],
        )

    def test_apply_fields(self):
        query_set = Mock()
        _fields = ['name', 'id', '-title']
        docs.BaseDocument.apply_fields(query_set, _fields)
        query_set.only.assert_called_once_with('name', 'id')
        query_set.only().exclude.assert_called_once_with('title')

    def test_apply_sort(self):
        query_set = Mock()
        docs.BaseDocument.apply_sort(query_set, ['name', 'id'])
        query_set.order_by.assert_called_once_with('name', 'id')

    def test_apply_sort_no_sort(self):
        query_set = Mock()
        docs.BaseDocument.apply_sort(query_set, [])
        assert not query_set.order_by.called

    def test_count(self):
        query_set = Mock()
        docs.BaseDocument.count(query_set)
        query_set.count.assert_called_once_with(
            with_limit_and_skip=True)

    def test_is_modified_no_changed_fields(self):
        obj = docs.BaseMixin()
        obj.pk_field = Mock(return_value='id')
        obj._get_changed_fields = Mock(return_value=[])
        obj.id = 1
        assert not obj._is_modified()

    def test_is_modified(self):
        obj = docs.BaseMixin()
        obj.pk_field = Mock(return_value='id')
        obj._get_changed_fields = Mock(return_value=['name'])
        obj.id = 1
        assert obj._is_modified()

    def test_is_created(self):
        obj = docs.BaseMixin()
        obj._created = True
        assert obj._is_created()
        obj._created = False
        assert not obj._is_created()


class TestBaseDocument(object):

    def test_init_created_with_invalid_fields(self):
        class MyModel(docs.BaseDocument):
            name = fields.StringField()

        with pytest.raises(FieldDoesNotExist):
            MyModel(name='foo', description='bar', id=1, pk=3)

    def test_init_loaded_with_invalid_fields(self):
        class MyModel(docs.BaseDocument):
            name = fields.StringField()

        try:
            MyModel(_created=False, name='foo', description='bar')
        except FieldDoesNotExist:
            raise Exception('Unexpected error')

    def test_get_null_values(self):
        class MyModel1(docs.BaseDocument):
            name = fields.StringField(primary_key=True)

        class MyModel2(docs.BaseDocument):
            name = fields.StringField(primary_key=True)
            models1 = fields.Relationship(
                document='MyModel1', backref_name='model2')

        assert MyModel1.get_null_values() == {
            'name': None,
            'model2': None,
        }

        assert MyModel2.get_null_values() == {
            'models1': [],
            'name': None,
        }

    def test_clean(self):
        class MyModel1(docs.BaseDocument):
            name = fields.IdField(onupdate='foo')
        obj = MyModel1()
        obj._created = False
        assert obj.name is None
        obj.clean()
        assert obj.name == 'foo'

    def test_to_dict(self):
        class MyModel1(docs.BaseDocument):
            id = fields.IdField()
            name = fields.StringField(primary_key=True)

        obj = MyModel1(name='foo')
        assert obj.to_dict() == {
            '_pk': 'foo',
            '_type': 'MyModel1',
            '_version': 0,
            'id': 'foo',
            'name': 'foo',
        }



from .fields import (RelationshipField, ReferenceField)

relationship_fields = (RelationshipField, ReferenceField)


def is_relationship_field(field, model_cls):
    """ Determine if `field` of the `model_cls` is a relational
    field.
    """
    if not model_cls.has_field(field):
        return False
    field_obj = model_cls._fields[field]
    return isinstance(field_obj, relationship_fields)


def get_relationship_cls(field, model_cls):
    """ Return class that is pointed to by relationship field
    `field` from model `model_cls`.

    Make sure field exists and is a relationship
    field manually. Use `is_relationship_field` for this.
    """
    field_obj = model_cls._fields[field]
    if isinstance(field_obj, RelationshipField):
        field_obj = getattr(field_obj, 'field')
    return getattr(field_obj, 'document_type')



import copy
import logging

import six
import mongoengine as mongo

from nefertari.json_httpexceptions import (
    JHTTPBadRequest, JHTTPNotFound, JHTTPConflict)
from nefertari.utils import (
    process_fields, process_limit, _split, dictset, drop_reserved_params)
from .metaclasses import ESMetaclass, DocumentMetaclass
from .signals import on_bulk_update
from .fields import (
    DateTimeField, IntegerField, ForeignKeyField, RelationshipField,
    DictField, ListField, ChoiceField, ReferenceField, StringField,
    TextField, UnicodeField, UnicodeTextField,
    IdField, BooleanField, BinaryField, DecimalField, FloatField,
    BigIntegerField, SmallIntegerField, IntervalField, DateField,
    TimeField, BaseFieldMixin
)


log = logging.getLogger(__name__)


def get_document_cls(name):
    try:
        return mongo.document.get_document(name)
    except:
        raise ValueError('`%s` does not exist in mongo db' % name)


def get_document_classes():
    """ Get all defined not abstract document classes

    Class is assumed to be non-abstract if its `_meta['abstract']` is
    defined and False.
    """
    document_classes = {}
    registry = mongo.base.common._document_registry.copy()
    for model_name, model_cls in registry.items():
        _meta = getattr(model_cls, '_meta', {})
        abstract = _meta.get('abstract', True)
        if not abstract:
            document_classes[model_name] = model_cls
    return document_classes


def process_lists(_dict):
    for k in _dict:
        new_k, _, _t = k.partition('__')
        if _t == 'in' or _t == 'all':
            _dict[k] = _dict.aslist(k)
    return _dict


def process_bools(_dict):
    for k in _dict:
        new_k, _, _t = k.partition('__')
        if _t == 'bool':
            _dict[new_k] = _dict.pop_bool_param(k)

    return _dict


TYPES_MAP = {
    StringField: {'type': 'string'},
    TextField: {'type': 'string'},
    UnicodeField: {'type': 'string'},
    UnicodeTextField: {'type': 'string'},
    mongo.fields.ObjectIdField: {'type': 'string'},
    ForeignKeyField: {'type': 'string'},
    IdField: {'type': 'string'},

    BooleanField: {'type': 'boolean'},
    BinaryField: {'type': 'object'},
    DictField: {'type': 'object', 'enabled': False},

    DecimalField: {'type': 'double'},
    FloatField: {'type': 'double'},

    IntegerField: {'type': 'long'},
    BigIntegerField: {'type': 'long'},
    SmallIntegerField: {'type': 'long'},
    IntervalField: {'type': 'long'},

    DateTimeField: {'type': 'date', 'format': 'dateOptionalTime'},
    DateField: {'type': 'date', 'format': 'dateOptionalTime'},
    TimeField: {'type': 'date', 'format': 'HH:mm:ss'},
}


class BaseMixin(object):
    """ Represents mixin class for models.

    Attributes:
        _auth_fields: String names of fields meant to be displayed to
            authenticated users.
        _public_fields: String names of fields meant to be displayed to
            non-authenticated users.
        _hidden_fields: String names of fields meant to be hidden but editable.
        _nested_relationships: String names of reference/relationship fields
            that should be included in JSON data of an object as full
            included documents. If reference/relationship field is not
            present in this list, this field's value in JSON will be an
            object's ID or list of IDs.
        _nesting_depth: Depth of relationship field nesting in JSON.
            Defaults to 1(one) which makes only one level of relationship
            nested.
    """
    _public_fields = None
    _auth_fields = None
    _hidden_fields = None
    _nested_relationships = ()
    _backref_hooks = ()
    _nesting_depth = 1

    _type = property(lambda self: self.__class__.__name__)
    Q = mongo.Q

    @classmethod
    def get_es_mapping(cls, _depth=None, types_map=None):
        """ Generate ES mapping from model schema. """
        from nefertari.elasticsearch import ES
        if types_map is None:
            types_map = TYPES_MAP
        if _depth is None:
            _depth = cls._nesting_depth
        depth_reached = _depth <= 0

        properties = {}
        mapping = {
            ES.src2type(cls.__name__): {
                'properties': properties
            }
        }
        fields = cls._fields.copy()
        for name, field in fields.items():
            if isinstance(field, RelationshipField):
                field = field.field
            if isinstance(field, (ReferenceField, RelationshipField)):
                if name in cls._nested_relationships and not depth_reached:
                    field_mapping = {'type': 'nested'}
                    submapping = field.document_type.get_es_mapping(
                        _depth=_depth-1)
                    field_mapping.update(list(submapping.values())[0])
                else:
                    field_mapping = types_map[
                        field.document_type.pk_field_type()]
                properties[name] = field_mapping
                continue

            if isinstance(field, ChoiceField):
                field = field._real_field
            field_type = type(field)
            if field_type is ListField:
                field_type = field.item_type
            if field_type not in types_map:
                continue
            properties[name] = types_map[field_type]

        properties['_pk'] = {'type': 'string'}
        return mapping

    @classmethod
    def autogenerate_for(cls, model, set_to):
        """ Setup `post_save` event handler.

        Event handler is registered for class :model: and creates a new
        instance of :cls: with a field :set_to: set to an instance on
        which the event occured.

        The handler is set up as class method because mongoengine refuses
        to call signal handlers if they aren't importable.
        """
        from mongoengine import signals

        def generate(cls, sender, document, *args, **kw):
            if kw.get('created', False):
                cls(**{set_to: document}).save()

        cls._generate_on_creation = classmethod(generate)
        signals.post_save.connect(cls._generate_on_creation, sender=model)

    @classmethod
    def pk_field(cls):
        return cls._meta['id_field']

    @classmethod
    def pk_field_type(cls):
        return getattr(cls, cls.pk_field()).__class__

    @classmethod
    def check_fields_allowed(cls, fields):
        if issubclass(cls, mongo.DynamicDocument):
            # Dont check if its dynamic doc
            return
        fields = [f.split('__')[0] for f in fields]
        fields_to_query = set(cls.fields_to_query())
        if not set(fields).issubset(fields_to_query):
            not_allowed = set(fields) - fields_to_query
            raise JHTTPBadRequest(
                "'%s' object does not have fields: %s" % (
                    cls.__name__, ', '.join(not_allowed)))

    @classmethod
    def filter_fields(cls, params):
        """ Filter out fields with invalid names. """
        fields = cls.fields_to_query()
        return dictset({
            name: val for name, val in params.items()
            if name.split('__')[0] in fields
        })

    @classmethod
    def apply_fields(cls, query_set, _fields):
        fields_only, fields_exclude = process_fields(_fields)

        try:
            if fields_only:
                query_set = query_set.only(*fields_only)

            if fields_exclude:
                query_set = query_set.exclude(*fields_exclude)

        except mongo.InvalidQueryError as e:
            raise JHTTPBadRequest('Bad _fields param: %s ' % e)

        return query_set

    @classmethod
    def apply_sort(cls, query_set, _sort):
        if not _sort:
            return query_set
        return query_set.order_by(*_sort)

    @classmethod
    def count(cls, query_set):
        return query_set.count(with_limit_and_skip=True)

    @classmethod
    def filter_objects(cls, objects, first=False, **params):
        """ Perform query with :params: on instances sequence :objects:

        Arguments:
            :object: Sequence of :cls: instances on which query should be run.
            :params: Query parameters to filter :objects:.
        """
        id_name = cls.pk_field()
        key = '{}__in'.format(id_name)
        ids = [getattr(obj, id_name, None) for obj in objects]
        ids = [str(id_) for id_ in ids if id_ is not None]
        params[key] = ids

        if first:
            return cls.get_item(**params)
        else:
            return cls.get_collection(**params)

    @classmethod
    def get_collection(cls, **params):
        """
        Params may include '_limit', '_page', '_sort', '_fields'.
        Returns paginated and sorted query set.
        Raises JHTTPBadRequest for bad values in params.
        """
        log.debug('Get collection: {}, {}'.format(cls.__name__, params))
        params.pop('__confirmation', False)
        _strict = params.pop('_strict', True)
        _item_request = params.pop('_item_request', False)

        _sort = _split(params.pop('_sort', []))
        _fields = _split(params.pop('_fields', []))
        _limit = params.pop('_limit', None)
        _page = params.pop('_page', None)
        _start = params.pop('_start', None)
        query_set = params.pop('query_set', None)

        _count = '_count' in params
        params.pop('_count', None)
        _explain = '_explain' in params
        params.pop('_explain', None)
        _raise_on_empty = params.pop('_raise_on_empty', False)

        if query_set is None:
            query_set = cls.objects

        # Remove any __ legacy instructions from this point on
        params = dictset({
            key: val for key, val in params.items()
            if not key.startswith('__')
        })

        params = drop_reserved_params(params)
        if _strict:
            _check_fields = [
                f.strip('-+') for f in list(params.keys()) + _fields + _sort]
            cls.check_fields_allowed(_check_fields)
        else:
            params = cls.filter_fields(params)

        process_lists(params)
        process_bools(params)

        # If param is _all then remove it
        params.pop_by_values('_all')

        try:
            query_set = query_set(**params)
            _total = query_set.count()
            if _count:
                return _total

            # Filtering by fields has to be the first thing to do on the
            # query_set!
            query_set = cls.apply_fields(query_set, _fields)
            query_set = cls.apply_sort(query_set, _sort)

            if _limit is not None:
                _start, _limit = process_limit(_start, _page, _limit)
                query_set = query_set[_start:_start+_limit]

            if not query_set.count():
                msg = "'%s(%s)' resource not found" % (cls.__name__, params)
                if _raise_on_empty:
                    raise JHTTPNotFound(msg)
                else:
                    log.debug(msg)

        except mongo.ValidationError as ex:
            if _item_request:
                msg = "'%s(%s)' resource not found" % (cls.__name__, params)
                raise JHTTPNotFound(msg, explanation=ex.message)
            else:
                raise JHTTPBadRequest(str(ex), extra={'data': ex})
        except mongo.InvalidQueryError as ex:
            raise JHTTPBadRequest(str(ex), extra={'data': ex})

        if _explain:
            return query_set.explain()

        log.debug('get_collection.query_set: %s(%s)',
                  cls.__name__, query_set._query)

        query_set._nefertari_meta = dict(
            total=_total,
            start=_start,
            fields=_fields)

        return query_set

    @classmethod
    def has_field(cls, field):
        return field in cls._fields

    @classmethod
    def fields_to_query(cls):
        query_fields = [
            'id', '_limit', '_page', '_sort', '_fields', '_count', '_start']
        return query_fields + list(cls._fields.keys())

    @classmethod
    def get_item(cls, **params):
        params.setdefault('_raise_on_empty', True)
        params['_limit'] = 1
        params['_item_request'] = True
        query_set = cls.get_collection(**params)
        return query_set.first()

    def unique_fields(self):
        pk_field = [self.pk_field()]
        uniques = [e['fields'][0][0] for e in self._unique_with_indexes()]
        return uniques + pk_field

    @classmethod
    def get_or_create(cls, **params):
        defaults = params.pop('defaults', {})
        try:
            return cls.objects.get(**params), False
        except mongo.queryset.DoesNotExist:
            defaults.update(params)
            return cls(**defaults).save(), True
        except mongo.queryset.MultipleObjectsReturned:
            raise JHTTPBadRequest('Bad or Insufficient Params')

    def _update(self, params, **kw):
        process_bools(params)
        self.check_fields_allowed(list(params.keys()))
        iter_fields = set(
            k for k, v in type(self)._fields.items()
            if isinstance(v, (DictField, ListField)) and
            not isinstance(v, RelationshipField))
        pk_field = self.pk_field()
        for key, value in params.items():
            if key == pk_field:  # can't change the primary key
                continue
            if key in iter_fields:
                self.update_iterables(value, key, unique=True, save=False)
            else:
                setattr(self, key, value)
        return self.save(**kw)

    @classmethod
    def _delete_many(cls, items, request=None):
        """ Delete objects from :items: """
        items_count = len(items)
        for item in items:
            item.delete(request)
        return items_count

    @classmethod
    def _update_many(cls, items, params, request=None):
        """ Update objects from :items:

        If :items: is an instance of `mongoengine.queryset.queryset.QuerySet`
        items.update() is called. Otherwise update is performed per-object.

        'on_bulk_update' is called explicitly, because mongoengine does not
        trigger any signals on QuerySet.update() call.
        """
        if isinstance(items, mongo.queryset.queryset.QuerySet):
            items.update(**params)
            on_bulk_update(cls, items, request)
            return cls.count(items)
        items_count = len(items)
        for item in items:
            item.update(params, request)
        return items_count

    def __repr__(self):
        parts = ['%s:' % self.__class__.__name__]

        pk_field = self.pk_field()
        parts.append('{}={}'.format(pk_field, getattr(self, pk_field)))

        if hasattr(self, '_version'):
            parts.append('v=%s' % self._version)

        return '<%s>' % ', '.join(parts)

    @classmethod
    def get_by_ids(cls, ids, **params):
        pk_field = '{}__in'.format(cls.pk_field())
        params.update({
            pk_field: ids,
        })
        return cls.get_collection(**params)

    @classmethod
    def get_null_values(cls):
        """ Get null values of :cls: fields. """
        skip_fields = {'_version', '_acl'}
        null_values = {}
        for name in cls._fields.keys():
            if name in skip_fields:
                continue
            field = getattr(cls, name)
            if isinstance(field, RelationshipField):
                value = []
            else:
                value = None
            null_values[name] = value
        null_values.pop('id', None)
        return null_values

    def to_dict(self, **kwargs):
        _depth = kwargs.get('_depth')
        if _depth is None:
            _depth = self._nesting_depth
        depth_reached = _depth is not None and _depth <= 0

        _data = dictset()
        for field, field_type in self._fields.items():
            # Ignore ForeignKeyField fields
            if isinstance(field_type, ForeignKeyField):
                continue
            value = getattr(self, field, None)

            if value is not None:
                include = field in self._nested_relationships
                if not include or depth_reached:
                    encoder = lambda v: getattr(v, v.pk_field(), None)
                else:
                    encoder = lambda v: v.to_dict(_depth=_depth-1)

                if isinstance(field_type, ReferenceField):
                    value = encoder(value)
                elif isinstance(field_type, RelationshipField):
                    value = [encoder(val) for val in value]
                elif hasattr(value, 'to_dict'):
                    value = value.to_dict(_depth=_depth-1)

            _data[field] = value
        _data['_type'] = self._type
        _data['_pk'] = str(getattr(self, self.pk_field()))
        return _data

    def get_related_documents(self, nested_only=False):
        """ Return pairs of (Model, istances) of relationship fields.

        Pair contains of two elements:
          :Model: Model class object(s) contained in field.
          :instances: Model class instance(s) contained in field

        :param nested_only: Boolean, defaults to False. When True, return
            results only contain data for models on which current model
            and field are nested.
        """
        relationship_fields = {
            name: field for name, field in self._fields.items()
            if isinstance(field, (ReferenceField, RelationshipField))}

        for name, field in relationship_fields.items():
            value = getattr(self, name)
            if not value:
                continue
            if not isinstance(value, list):
                value = [value]
            model_cls = value[0].__class__

            if nested_only:
                backref = getattr(field, 'reverse_rel_field', None)
                if backref and backref not in model_cls._nested_relationships:
                    continue

            yield (model_cls, value)

    def update_iterables(self, params, attr, unique=False,
                         value_type=None, save=True,
                         request=None):
        is_dict = isinstance(type(self)._fields[attr], mongo.DictField)
        is_list = isinstance(type(self)._fields[attr], mongo.ListField)

        def split_keys(keys):
            neg_keys = []
            pos_keys = []

            for key in keys:
                if key.startswith('__'):
                    continue
                if key.startswith('-'):
                    neg_keys.append(key[1:])
                else:
                    pos_keys.append(key.strip())
            return pos_keys, neg_keys

        def update_dict(update_params):
            final_value = getattr(self, attr, {}) or {}
            final_value = final_value.copy()
            if update_params is None or update_params == '':
                if not final_value:
                    return
                update_params = {
                    '-' + key: val for key, val in final_value.items()}
            positive, negative = split_keys(list(update_params.keys()))

            # Pop negative keys
            for key in negative:
                final_value.pop(key, None)

            # Set positive keys
            for key in positive:
                final_value[str(key)] = update_params[key]

            setattr(self, attr, final_value)
            if save:
                self.save(request)

        def update_list(update_params):
            final_value = getattr(self, attr, []) or []
            final_value = copy.deepcopy(final_value)
            if update_params is None or update_params == '':
                if not final_value:
                    return
                update_params = ['-' + val for val in final_value]
            if isinstance(update_params, dict):
                keys = list(update_params.keys())
            else:
                keys = update_params

            positive, negative = split_keys(keys)

            if not (positive + negative):
                raise JHTTPBadRequest('Missing params')

            if positive:
                if unique:
                    positive = [v for v in positive if v not in final_value]
                final_value += positive

            if negative:
                final_value = list(set(final_value) - set(negative))

            setattr(self, attr, final_value)
            if save:
                self.save(request)

        if is_dict:
            update_dict(params)

        elif is_list:
            update_list(params)

    @classmethod
    def expand_with(cls, with_cls, join_on=None, attr_name=None, params={},
                    with_params={}):
        """ Acts like "join" and inserts the with_cls objects in
        the result as "attr_name".
        """
        if join_on is None:
            join_on = with_cls.__name__.lower()
        if attr_name is None:
            attr_name = with_cls.__name__.lower()

        with_params.pop('_fields', [])
        with_objs = with_cls.get_by_ids(
            cls.objects.scalar(join_on),
            **with_params)
        with_objs = dict([[str(wth.id), wth] for wth in with_objs])

        params['%s__in' % join_on] = list(with_objs.keys())
        objs = cls.get_collection(**params)

        for ob in objs:
            ob._data[attr_name] = with_objs[getattr(ob, join_on)]
            setattr(ob, attr_name, ob._data[attr_name])

        return objs

    def _is_modified(self):
        """ Determine if instance is modified.

        For instance to be marked as 'modified', it should:
          * Have PK field set (not newly created)
          * Have changed fields
        """
        modified = bool(self._get_changed_fields())
        return modified

    def _is_created(self):
        return self._created

    def _to_python_fields(self):
        """ Call to_python on non-relation fields. """
        from .utils import relationship_fields
        for name, field in self._fields.items():
            rel_field = isinstance(field, relationship_fields)
            if name not in self._data or rel_field:
                continue
            value = self._data[name]
            try:
                self._data[name] = field.to_python(value)
            except:
                continue


class BaseDocument(six.with_metaclass(DocumentMetaclass,
                                      BaseMixin, mongo.Document)):
    _version = IntegerField(default=0)

    meta = {
        'abstract': True,
    }

    def __init__(self, *args, **values):
        """ Override init to filter out invalid fields from :values:.

        Fields are filtered out to make mongoengine less strict when
        loading objects from database.
        :internal_fields: are the fields pop'ed from :values: before
        performing fields presence validation in the original mongoengine
        init code:
        https://github.com/MongoEngine/mongoengine/blob/v0.9.0/mongoengine/base/document.py#L41

        PS. This issue is fixed in mongoengine master and not released after
        0.9.0 yet.
        https://github.com/MongoEngine/mongoengine/blob/master/mongoengine/base/document.py#L75
        """
        _created = values.get('_created')
        if _created is not None and not _created:
            internal_fields = [
                'id', 'pk', '_cls', '_text_score',
                '__auto_convert', '__only_fields', '_created',
            ]
            valid_fields = list(self._fields.keys()) + internal_fields
            values = {key: val for key, val in values.items()
                      if key in valid_fields}
        super(BaseDocument, self).__init__(*args, **values)

    def _bump_version(self):
        if self._is_modified():
            self._version += 1

    def save(self, request=None, *arg, **kw):
        """
        Force insert document in creation so that unique constraits are
        respected.
        This makes each POST to a collection act as a 'create' operation
        (as opposed to an 'update' for example).
        """
        kw['force_insert'] = self._created
        self._request = request
        self._bump_version()
        try:
            super(BaseDocument, self).save(*arg, **kw)
        except (mongo.NotUniqueError, mongo.OperationError) as e:
            if (e.__class__ is mongo.OperationError
                    and 'E11000' not in e.message):
                raise  # Other error, not duplicate

            raise JHTTPConflict(
                detail='Resource `{}` already exists.'.format(
                    self.__class__.__name__),
                extra={'data': e})
        else:
            self.run_backref_hooks()
            self._backref_hooks = ()
            return self

    def run_backref_hooks(self):
        """ Runs post-save backref hooks.

        Includes backref hooks which are used one time only
        to sync the backrefs.
        """
        for hook in self._backref_hooks:
            hook(document=self)

    def update(self, params, request=None, **kw):
        kw['request'] = request
        # request are passed to _update and then to save
        try:
            self._update(params, **kw)
            self._to_python_fields()
            return self
        except (mongo.NotUniqueError, mongo.OperationError) as e:
            if (e.__class__ is mongo.OperationError
                    and 'E11000' not in e.message):
                raise  # other error, not duplicate

            raise JHTTPConflict(
                detail='Resource `{}` already exists.'.format(
                    self.__class__.__name__),
                extra={'data': e})

    def validate(self, *arg, **kw):
        try:
            super(BaseDocument, self).validate(*arg, **kw)
        except mongo.ValidationError as e:
            raise JHTTPBadRequest(
                'Resource `%s`: %s' % (self.__class__.__name__, e),
                extra={'data': e})

    def delete(self, request=None, **kw):
        self._request = request
        super(BaseDocument, self).delete(**kw)

    @classmethod
    def get_field_params(cls, field_name):
        """ Get init params of field named :field_name:. """
        field = cls._fields.get(field_name)
        return getattr(field, '_init_kwargs', None)

    def clean(self):
        """ Clean fields which are instances of BaseFieldMixin """
        for field_name, field_obj in self._fields.items():
            if isinstance(field_obj, BaseFieldMixin):
                field_obj.clean(self)


class ESBaseDocument(six.with_metaclass(ESMetaclass, BaseDocument)):
    """ Base for document classes which should be indexed by ES. """
    meta = {
        'abstract': True,
    }



from setuptools import setup, find_packages

install_requires = [
    'elasticsearch',
    'mongoengine==0.9',
    'nefertari>=0.6.1',
    'pymongo==2.8',
    'pyramid_tm',
    'python-dateutil',
    'six',
    'zope.dottedname',
]

setup(
    name='nefertari_mongodb',
    version="0.4.1",
    description='mongodb engine for nefertari',
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
    ],
    author='Brandicted',
    author_email='hello@brandicted.com',
    url='https://github.com/brandicted/nefertari-mongodb',
    keywords='web wsgi bfg pylons pyramid rest mongodb mongoengine',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
)
